open Algebra
open Assertions

(******************************************************************************)
(* Implementing the Ordered interface with ints *******************************)
(******************************************************************************)

(* int implementation of Ordered interface.  See .mli *)
module Int = struct
  module Core = struct
    (* Abstraction function: the int n is interpreted as the integer n *)
    (* Rep invariant: none; all ints are valid *)
    type t = int

    (* since all ints are valid, repOk never fails! *)
    let repOk x = x

    let (=) a b = a = b
    let (+) a b = repOk (a + b)
    let (~-) a  = repOk (-a)
    let zero    = repOk (0)

    let ( * ) a b = repOk (a * b)
    let one = repOk (1)

    let isPositive a = a > 0
  end
  include Core
  include ExtendOrdered(Core)
end

(* test all of the Ordered properties for Int.  See "Note on testing" in the summary
   file for details on this.  *)
TEST_MODULE "Ints" = OrderedSpec(Int)(struct
  let examples = [1;2;3;4;-1;max_int;min_int]
end)

(* floats.  See .mli *)
module Float = struct
  module Core = struct
    (* Abstraction function:
         We interpret floating point numbers as real numbers in the usual way
         (see the section "Floating Point Arithmetic" in the Pervasives module
         for more details..
       Rep invariant:
         The special numbers nan, infinity, and neg_infinity are not considered
         valid Float.ts.
    *)
    type t = float

    let repOk x = match classify_float x with
      | FP_infinite | FP_nan -> failwith "Float Rep invariant failed"
      | _                    -> x

    let (=) a b = (a = b)
    let (+) a b = repOk (a +. b)
    let (~-) a  = repOk (-.a)
    let zero    = repOk (0.)

    let ( * ) a b = repOk (a *. b)
    let one   = repOk (1.)

    let isPositive a = a > 0.

    let inv a = repOk (1. /. a)
  end

  include Core
  include ExtendNumber(Core)
end

(* test all the Number properties for Float.  See "Note on testing" in lecture
   summary. *)
TEST_MODULE "Floats" = NumberSpec(Float)(struct
  let examples = [-1.; 1.; 0.; 10000.; -10000.; 0.00001; max_float]
end)

(* big integers.  See .mli

   Note: to compile this, you need to include the ocaml nums library.  Be sure
   that the attached .cs3110 file is in your current directory if compiling with
   the cs3110 tools.

   See the OCaml BigInt module for a list of big_int functions
*)
module BigInt = struct
  module Core = struct
    open Big_int

    (* abstraction function: a t represents an integer using the same
         interpretation as the Big_int module's.
       representation invariant: none!  Every big_int is valid.
    *)
    type t = big_int

    let (+)   = add_big_int
    let (~-)  = minus_big_int
    let ( * ) = mult_big_int
    let (=)   = eq_big_int

    let one  = unit_big_int
    let zero = zero_big_int

    let isPositive n = (sign_big_int n) > 0
  end

  include Core
  include ExtendOrdered(Core)
end

(* Test all of the Ordered properties for BigInt *)
TEST_MODULE "BigInts" = OrderedSpec(BigInt)(struct
  let examples = List.map Big_int.big_int_of_int [1;2;3;4;-1;max_int;min_int]
end)

(******************************************************************************)
(** The Fraction functor ******************************************************)
(******************************************************************************)

(* see .mli *)
module type Fraction = sig
  module Numerator : Multipliable
  include Dividable

  val from_numerator : Numerator.t -> t
  val to_pair        : t -> Numerator.t * Numerator.t
end

(** Fraction functor implementation *)
module Fraction (M : Multipliable) = struct
  module Numerator = M

  module Core = struct
    (* Abstraction function: The pair (n,d) represents the rational number n/d.
       Rep invariant: The denominator d is non-zero *)
    type t = M.t * M.t

    let repOk (n,d) = if M.(d = zero) then failwith "Fraction Rep. invariant failed"
                                      else (n,d)

    let (=)   (n1,d1) (n2,d2) =       M.(n1*d2 = n2*d1)
    let (+)   (n1,d1) (n2,d2) = repOk M.(n1*d2 + n2*d1, d1*d2)
    let (~-)  (n1,d1)         = repOk M.(-n1, d1)
    let zero                  = repOk M.(zero, one)
    let one                   = repOk M.(one,  one)
    let ( * ) (n1,d1) (n2,d2) = repOk M.(n1*n2, d1*d2)
    let inv (n, d) = if M.(n = zero) then failwith "Division by zero!"
                                     else repOk (d,n)

  end

  include Core
  include ExtendDividable(Core)

  let to_pair f = f
  let from_numerator n = (n, M.one)
end

(* see .mli *)
module type OrdFraction = sig
  include Fraction
  include Ordered with type t := t
end

module OrdFraction (M : Ordered) = struct
  module OCore = struct
    module FC = Fraction(M)
    include FC.Core

    open M
    let isPositive (n,d) =
      if n = zero then false
      else if isPositive n then isPositive d
      else not (isPositive d)
  end

  include Fraction(M)
  include ExtendOrdered(OCore)

end

(******************************************************************************)
(* The Vector type ************************************************************)
(******************************************************************************)

(* See .mli *)
module type Vector2 = sig

  module Scalar : Dividable

  include Addable

  (* additional functions that Vectors support that general Addables do not *)
  val dot   : t -> t -> Scalar.t
  val cross : t -> t
  val norm2 : t -> Scalar.t
  val smult : Scalar.t -> t -> t
  val sdiv  : t -> Scalar.t -> t
end

(* See .mli *)
module Vector2 (D : Dividable) = struct
  module Scalar = D

  module Core = struct
    (* abstraction function: we interpret (x,y) as the vector (x,y)
       rep invariant: none!  all pairs are valid. *)
    type t = D.t * D.t

    let repOk x = x

    let (+)  (x1,y1) (x2,y2) = repOk D.(x1+x2, y1+y2)
    let (~-) (x,y)           = repOk D.(-x, -y)
    let zero                 = repOk D.(zero, zero)
    let (=)  (x1,y1) (x2,y2) =       D.(x1 = x2 && y1 = y2)
  end

  include Core
  include ExtendAddable(Core)

  let dot   (x1,y1) (x2,y2) = D.(x1*x2 + y1*y2)
  let cross (x,y)           = repOk D.(-y,x)
  let norm2 v               = dot v v
  let smult s (x,y)         = repOk D.(s*x,s*y)
  let sdiv  (x,y) s         = repOk D.(x/s,s/y)
end

(** We define a few useful modules from the functors we have built *)

module Rational     = OrdFraction(Int)
module BigRational  = OrdFraction(BigInt)
module Vec2OfFloat  = Vector2 (Float)
module Vec2OfBigRat = Vector2 (BigRational)
module Vec2OfRat    = Vector2 (Rational)

(** and we test them all.  See "Note on testing" in the lecture summary. *)

(** Note: although the .mli makes the concrete types defined in these modules
    opaque, the entire .ml file is type checked without reference to the .mli,
    so within this file, the types are not opaque.  This allows us to write the
    examples below and use them as "t"s of the various modules. *)

(* test Rational *)
let frac_examples = [(1,2); (1,1); (2,2); (0,1); (-1,1); (max_int,1); (1,max_int); (0,min_int+1)]

TEST_MODULE "Rationals" = NumberSpec(Rational)(struct
  let examples = frac_examples
end)

(* test BigRational *)
TEST_MODULE "BigRationals" = NumberSpec(BigRational)(struct
  let examples = 
    List.map (fun (n,d) -> Big_int.big_int_of_int n, Big_int.big_int_of_int d)
              frac_examples
end)

(* test Vec2OfFloat *)
TEST_MODULE "Vec2OfFloat" = AddableSpec(Vec2OfFloat)(struct
  let examples = [(0.,0.); (0.,1.); (1.,0.); (1.,1.); (max_float,max_float); (-1.,-1.);]
end)


(* test Vec2OfRat *)
let frac_vec_examples = [(0,1),(0,1); (1,1),(0,1); (1,1),(0,1); (max_int-1,max_int-1),(0,1)]

TEST_MODULE "Vec2OfRat" = AddableSpec(Vec2OfRat)(struct
  let examples = frac_vec_examples
end)

let frac_vec_big_examples =
  let bi = Big_int.big_int_of_int in
  let to_big_vec ((nx,dx),(ny,dy)) = (bi nx, bi dx), (bi ny, bi dy) in
  List.map to_big_vec frac_vec_examples

TEST_MODULE "Vec2OfBigRat" = AddableSpec(Vec2OfBigRat)(struct
  let examples = frac_vec_big_examples
end)
